import json
from typing import Any

from ipykernel import zmqshell  # type: ignore
from IPython.core import magic  # type: ignore

from repltilian import SwiftREPL, repl_output


@magic.magics_class
class REPLMagic(magic.Magics):  # type: ignore
    def __init__(self, shell: zmqshell.ZMQInteractiveShell):
        """Initialize the REPLMagic class state."""
        super().__init__(shell)
        self.flip = False
        self._repl_instance: SwiftREPL | None = None

    def get_repl(self) -> SwiftREPL:
        if self._repl_instance is None:
            raise ValueError(
                "REPL is not initialized. Use %repl_init path/package to initialize the REPL."
            )
        return self._repl_instance

    @magic.line_magic  # type: ignore
    def repl_init(self, line: str) -> None:
        """Initialize (or reinitialize) the REPL instance.
        Usage: %repl_init path/to/package
        """
        line = line.strip()
        if self._repl_instance is not None:
            print("Closing previous REPL instance.")
            self._repl_instance.close()
        if line:
            print(f"Initializing REPL with package: '{line}' ...")
        else:
            print("Initializing REPL ...")
        self._repl_instance = SwiftREPL(line or None)

    @magic.cell_magic  # type: ignore
    def repl(self, line: str, cell: str) -> None:
        """Run the code in the REPL.

        Parameters:
            --verbose: print the output of the REPL
            --autoreload: reload the REPL instance
        """
        verbose = "verbose" in line
        autoreload = "autoreload" in line
        self.get_repl().run(cell, autoreload=autoreload, verbose=verbose)

    @magic.line_magic  # type: ignore
    def repl_instance(self, line: str) -> SwiftREPL:
        """Get the current REPL instance."""
        return self.get_repl()

    @magic.line_magic  # type: ignore
    def repl_add_file(self, line: str) -> None:
        """Add the file to the list of files which are reloaded before running the code.
        Usage: %repl_add_file path/to/file
        """
        line = line.strip()
        self.get_repl().add_reload_file(line)

    @magic.line_magic  # type: ignore
    def repl_get(self, line: str) -> Any:
        """Get the value of the variable from the REPL.
        Usage: %repl_get variable_name
        """
        line = line.strip()
        return self.get_repl().vars[line].get()

    @magic.line_magic  # type: ignore
    def repl_set(self, line: str) -> None:
        """Set the value of the variable in the running REPL.
        Usage: %repl_set var_name: var_type = var_value
        """
        repl = self.get_repl()
        line = line.strip()

        variables = repl_output.find_variables(line)
        if len(variables) != 1:
            print(
                "Invalid variable declaration! Expected format: " "var_name: var_type = var_value"
            )
            return
        var_name = list(variables.keys())[0]
        var_type, var_value = variables[var_name]
        var_value = json.loads(var_value)
        repl.vars.set(var_name, var_type, var_value)



from pathlib import Path

import pytest

import repltilian

THIS_FILE_DIR = Path(__file__).parent
RESOURCES_DIR = THIS_FILE_DIR / "resources"


@pytest.fixture()
def sample_filepath() -> str:
    return str(RESOURCES_DIR / "demo.swift")


@pytest.fixture()
def sample_code() -> str:
    with open(RESOURCES_DIR / "demo.swift") as f:
        return f.read()


@pytest.fixture()
def repl() -> repltilian.SwiftREPL:
    return repltilian.SwiftREPL()



"""Functions related to parsing Swift REPL output."""
import re

from repltilian import constants

# var_name: var_type = var_value or $R\d: var_type = var_value
VARIABLE_LINE_PATTERN = r"^(\w+|\$R+\d):\s*(.*?)\s*=\s*(.*)$"


def clean(text: str) -> str:
    """Cleans the raw output from a Swift REPL prompt output."""
    # Initialize the screen buffer as a list of lines
    screen_lines = [""]
    line = 0  # Current line index
    col = 0  # Current column index
    i = 0  # Current index in the input text
    n = len(text)

    def read_escape_sequence(s: str, i_: int) -> tuple[str | None, int]:
        """Reads an ANSI escape sequence starting from index i in string s.

        Returns the escape sequence and the index after the sequence.
        """
        if s[i_ : i_ + 2] != "\x1b[":
            return None, i_ + 1  # Not an escape sequence we recognize
        j = i_ + 2
        while j < len(s):
            if "A" <= s[j] <= "Z" or "a" <= s[j] <= "z":
                # End of the escape sequence
                j += 1  # Include the command character
                return s[i_:j], j
            else:
                j += 1
        # If we reach here, didn't find the end of escape sequence
        return s[i_:j], j

    while i < n:
        c = text[i]
        if c == "\x1b":
            # Process escape sequence
            esc_seq, next_i = read_escape_sequence(text, i)
            if esc_seq:
                # Process the escape sequence
                i = next_i
                # Extract the command and parameters
                cmd_match = re.match(r"\x1b\[([0-9;]*)([A-Za-z])", esc_seq)
                if cmd_match:
                    params = cmd_match.group(1)
                    cmd = cmd_match.group(2)
                    # Process the command
                    if cmd == "G":
                        # Cursor horizontal absolute
                        # Move cursor to column N (default 1)
                        N = int(params) if params else 1
                        col = max(N - 1, 0)  # Adjust for zero-based index
                    elif cmd == "J":
                        # Erase display from cursor to end of screen
                        # Clear from current line and position to end
                        screen_lines = screen_lines[: line + 1]
                        # Truncate the current line from cursor position
                        screen_lines[line] = screen_lines[line][:col]
                    else:
                        # Other commands can be implemented as needed
                        pass
                continue
            else:
                # Not a recognized escape sequence
                i += 1
                continue
        elif c == "\r":
            # Carriage return
            col = 0
            i += 1
        elif c == "\n":
            # Line feed
            line += 1
            if line >= len(screen_lines):
                screen_lines.append("")
            i += 1
        else:
            # Printable character
            # Ensure the current line exists
            while len(screen_lines) <= line:
                screen_lines.append("")
            current_line = screen_lines[line]
            # Extend the line with spaces if necessary
            if col > len(current_line):
                current_line += " " * (col - len(current_line))
            # Insert or replace the character at the current position
            if col < len(current_line):
                current_line = current_line[:col] + c + current_line[col + 1 :]
            else:
                current_line += c
            # Update the screen buffer
            screen_lines[line] = current_line
            # Move cursor forward
            col += 1
            i += 1

    # After processing, join the lines
    result = "\n".join(screen_lines)
    return result.strip()


def search_for_error(cleaned_output: str) -> str | None:
    """Check if the text contains an error message: "error:" or "$E{number}:"."""
    error_pattern = re.compile(r"(\$E\d+):|^error:")
    for line in cleaned_output.split("\n"):
        if error_pattern.search(line):
            return line
    return None


def _find_end_of_include_line(cleaned_output_lines: list[str]) -> int:
    for i, line in enumerate(cleaned_output_lines):
        if constants.END_OF_INCLUDE in line:
            return i
    return -1


def _split_output_by_end_of_include(cleaned_output: str) -> tuple[str, str]:
    cleaned_output_lines = cleaned_output.split("\n")
    end_of_include_line = _find_end_of_include_line(cleaned_output_lines)
    if end_of_include_line == -1:
        return "", cleaned_output
    include_part = "\n".join(cleaned_output_lines[:end_of_include_line])
    code_part = "\n".join(cleaned_output_lines[end_of_include_line + 1 :])
    return include_part, code_part


def remove_prompt_input_lines(cleaned_output: str) -> str:
    r"""Searches for lines which start from \d+[>.] and removes them."""
    lines = cleaned_output.split("\n")
    stop = False
    i = 0
    for i, line in enumerate(lines):
        if re.match(r"^\d+[>.]", line.strip()):
            stop = True
        elif stop:
            break
    return "\n".join(lines[i:])


def remove_prompt_variables_lines(cleaned_output: str) -> str:
    r"""Searches for lines which start from \d+[>.] and removes them."""
    lines = cleaned_output.split("\n")
    stop = False
    i = 0
    for i, line in enumerate(lines):
        if re.match(VARIABLE_LINE_PATTERN, line.strip()):
            stop = True
        elif stop:
            break
    return "\n".join(lines[: i - 1])


def print_output(
    cleaned_output: str,
    stop_output_at_pattern: str | None = None,
    hide_inputs: bool = False,
    hide_variables: bool = False,
) -> None:
    """Print the cleaned output from REPL. Optionally limit the output by a stop pattern.

    Args:
        cleaned_output: swift REPL cleaned output
        stop_output_at_pattern: pattern to stop the output rendering
        hide_inputs: hide input prompt lines from REPL output
        hide_variables: hide variable declarations from REPL output
    """
    _, output = _split_output_by_end_of_include(cleaned_output)
    if hide_inputs:
        output = remove_prompt_input_lines(output)

    if hide_variables:
        output = remove_prompt_variables_lines(output)

    if stop_output_at_pattern is not None:
        search_pattern = re.compile(stop_output_at_pattern)
        output_lines = output.split("\n")
        stop_k = 0
        for i, line in enumerate(output_lines):
            stop_k = i
            if search_pattern.search(line):
                break
        output = "\n".join(output_lines[:stop_k])

    output_lines = [line for line in output.split("\n") if line]
    output = "\n".join(output_lines).strip()
    if output:
        print(output)


def find_variables(cleaned_output: str) -> dict[str, tuple[str, str]]:
    """Extract variable declarations from the cleaned REPL output. This function searches for
    lines with the following pattern: "var_name: var_type = var_value" and returns a dictionary
    with the variable name as key and a tuple with the variable type and value content as string.

    Note: that for very long outputs not all variables may be captured as the REPL output is
    truncated.

    Args:
        cleaned_output: cleaned output from the REPL
    """
    register = {}
    lines = cleaned_output.splitlines()
    variable_pattern = re.compile(VARIABLE_LINE_PATTERN)

    in_value = False
    brace_counter = 0
    current_var_name = ""
    current_var_type = ""
    current_var_value = ""

    for line in lines:
        line = line.strip()
        if line.startswith("warning:") or line.startswith("error:"):
            continue
        if not in_value:
            m = re.match(variable_pattern, line)
            if m:
                var_name: str = m.group(1).strip()
                var_type: str = m.group(2).strip()
                var_value: str = m.group(3).strip()
                # Check for opening and closing braces in the initial value
                open_braces = var_value.count("{")
                close_braces = var_value.count("}")
                brace_counter = open_braces - close_braces
                if brace_counter > 0:
                    # Value spans multiple lines
                    in_value = True
                    current_var_value = var_value + "\n"  # Include the newline
                    current_var_name = var_name
                    current_var_type = var_type
                else:
                    # Value is on a single line
                    register[var_name] = (var_type, var_value)
            else:
                # Not a variable declaration line
                continue
        else:
            # Collecting multi-line variable value
            current_var_value += line + "\n"
            brace_counter += line.count("{") - line.count("}")
            if brace_counter <= 0:
                # Finished collecting the variable value
                register[current_var_name] = (current_var_type, current_var_value)
                # Reset for the next variable
                in_value = False
                current_var_name = ""
                current_var_type = ""
                current_var_value = ""
                brace_counter = 0

    return register


def batch_prompt(prompt: str, maxsize: int = 50) -> list[str]:
    blocks: list[str] = []
    current_block: list[str] = []
    for line in prompt.split("\n"):
        if not line.strip():
            # skip empty lines
            continue
        if line.strip().startswith("//"):
            # skip comments lines
            continue
        current_size = len("\n".join(current_block + [line]))
        if current_size >= maxsize:
            blocks.append("\n".join(current_block))
            current_block = [line]
        else:
            current_block.append(line)

    if current_block:
        blocks.append("\n".join(current_block))

    return blocks


def split_prompt(prompt: str, maxsize: int = 1000) -> list[str]:
    blocks: list[str] = []
    pos = 0
    while True:
        block = prompt[pos: pos + maxsize]
        blocks.append(block)
        pos += maxsize
        if pos >= len(prompt):
            break

    return blocks



from repltilian import repl_output


def assert_lines_equal(left: str, right: str) -> None:
    left = left.strip()
    right = right.strip()
    for i, (ll, rl) in enumerate(zip(left, right)):
        assert ll.strip() == rl.strip(), f"Line {i}: {ll} != {rl} are not equal"


def test__clean_prompt1() -> None:
    """Example 1.
    var point = Point<Float>(x: 1, y: 2)
    """
    repl_output_str = (
        "\x1b[1G\x1b[1G\x1b[J 63>  \r\n 64.  \x1b[1G 64. \r 64. \x1b[1G 64. \x1b["
        "6Gvar point = Point<Float>(x\r 64. var point = Point<Float>(x:\x1b[1G 64. \x1b[33G 1, "
        "y\r 64. var point = Point<Float>(x: 1, y:\x1b[1G 64. \x1b[39G 2)\x1b[1G\x1b[1G\x1b[J 64. "
        "var point = Point<Float>(x: 1, y: 2) \r\n 65.  \x1b[1G 65. \r 65. \x1b[1G 65. \x1b["
        "6G\x1b[6G\r\npoint: Point<Float> = {\r\n  x = 1\r\n  y = 2\r\n}\r\n\x1b[1G\x1b[J 65>  "
        "\x1b[1G 65> \r 65> \x1b[1G 65> \x1b[6G"
    )
    cleaned_output = repl_output.clean(repl_output_str)
    expected_output = (
        "63>  \n 64. var point = Point<Float>(x: 1, y: 2) \n 65.  \npoint: "
        "Point<Float> = {\n  x = 1\n  y = 2\n}\n 65>"
    )  # noqa: E501
    assert cleaned_output == expected_output


def test__clean_prompt2() -> None:
    """Example 2. var i = 1.0."""
    repl_output_str = (
        "var i = 1.0\x1b[18G\r\ni: Double = 1\r\n\x1b[1G\x1b[J 110>  \x1b[1G 110> "
        "\r 110> \x1b[1G 110> \x1b[7G"
    )  # noqa: E501
    cleaned_output = repl_output.clean(repl_output_str)
    expected_output = "var i = 1.0\ni: Double = 1\n 110>"
    assert cleaned_output == expected_output


def test__clean_prompt3() -> None:
    """Example 3.
    //cpp
    var point1 = Point<Float>(x: 5, y: 2)
    var point2 = Point<Float>(x: 5, y: 2)
    let result = point1 + point2 + point
    result.translate(dx: 1, dy: 100)
    """
    repl_output_str = (
        "//cpp\x1b[1G\x1b[1G\x1b[J 115> //cpp \r\n 116.  \x1b[1G 116. \r 116. \x1b[1G 116. "
        "\x1b[7Gvar point1 = Point<Float>(x\r 116. var point1 = Point<Float>(x:\x1b[1G 116. "
        "\x1b[35G 5, y\r 116. var point1 = Point<Float>(x: 5, y:\x1b[1G 116. \x1b[41G 2)\x1b["
        "1G\x1b[1G\x1b[J 116. var point1 = Point<Float>(x: 5, y: 2) \r\n 117.  \x1b[1G 117. \r "
        "117. \x1b[1G 117. \x1b[7Gvar point2 = Point<Float>(x\r 117. var point2 = Point<Float>("
        "x:\x1b[1G 117. \x1b[35G 5, y\r 117. var point2 = Point<Float>(x: 5, y:\x1b[1G 117. "
        "\x1b[41G 2)\x1b[1G\x1b[1G\x1b[J 117. var point2 = Point<Float>(x: 5, y: 2) \r\n 118.  "
        "\x1b[1G 118. \r 118. \x1b[1G 118. \x1b[7Glet result = point1 + point2 + point\x1b["
        "1G\x1b[1G\x1b[J 118. let result = point1 + point2 + point \r\n 119.  \x1b[1G 119. \r "
        "119. \x1b[1G 119. \x1b[7Gresult.translate(dx\r 119. result.translate(dx:\x1b[1G 119. "
        "\x1b[27G 1, dy\r 119. result.translate(dx: 1, dy:\x1b[1G 119. \x1b[34G 100)\x1b[1G\x1b["
        "1G\x1b[J 119. result.translate(dx: 1, dy: 100) \r\n 120.  \x1b[1G 120. \r 120. \x1b[1G "
        "120. \x1b[7G\x1b[7G\r\n$R1: Point<Float> = {\r\n  x = 12\r\n  y = 106\r\n}\r\npoint1: "
        "Point<Float> = {\r\n  x = 5\r\n  y = 2\r\n}\r\npoint2: Point<Float> = {\r\n  x = 5\r\n  y "
        "= 2\r\n}\r\nresult: Point<Float> = {\r\n  x = 11\r\n  y = 6\r\n}\r\n\x1b[1G\x1b[J 120>  "
        "\x1b[1G 120> \r 120> \x1b[1G 120> \x1b[7G"
    )
    cleaned_output = repl_output.clean(repl_output_str)
    expected_output = (
        "115> //cpp \n 116. var point1 = Point<Float>(x: 5, y: 2) \n 117. var "
        "point2 = Point<Float>(x: 5, y: 2) \n 118. let result = point1 + point2 + "
        "point \n 119. result.translate(dx: 1, dy: 100) \n 120.  \n$R1: "
        "Point<Float> = {\n  x = 12\n  y = 106\n}\npoint1: Point<Float> = {\n  x = 5\n  y = "
        "2\n}\npoint2: Point<Float> = {\n  x = 5\n  y = 2\n}\nresult: Point<Float> = {\n  x = "
        "11\n  y = 6\n}\n 120>"
    )
    assert cleaned_output == expected_output


def test__clean_prompt_with_error_in_output() -> None:
    """Example with error in output.
    //cpp
    result.error(dx: 1, dy: 100)
    """
    repl_output_str = (
        "//cpp\x1b[1G\x1b[1G\x1b[J 130> //cpp \r\n 131.  \x1b[1G 131. \r 131. \x1b["
        "1G 131. \x1b[7Gresult.error(dx\r 131. result.error(dx:\x1b[1G 131. \x1b["
        "23G 1, dy\r 131. result.error(dx: 1, dy:\x1b[1G 131. \x1b[30G 100)\x1b["
        "1G\x1b[1G\x1b[J 131. result.error(dx: 1, dy: 100) \r\n 132.  \x1b[1G 132. "
        "\r 132. \x1b[1G 132. \x1b[7G\x1b[7G\r\nerror: repl.swift:131:8: value of "
        "type 'Point<Float>' has no member 'error'\r\nresult.error(dx: 1, "
        "dy: 100)\r\n~~~~~~ ^~~~~\r\n\r\n\r\n\x1b[1G\x1b[J 130>  \x1b[1G 130> \r "
        "130> \x1b[1G 130> \x1b[7G"
    )
    cleaned_output = repl_output.clean(repl_output_str)
    expected_output = (
        "130> //cpp \n 131. result.error(dx: 1, dy: 100) \n 132.  \nerror: "
        "repl.swift:131:8: value of type 'Point<Float>' has no member "
        "'error'\nresult.error(dx: 1, dy: 100)\n~~~~~~ ^~~~~\n\n\n 130>"
    )
    assert cleaned_output == expected_output


def test__has_error() -> None:
    text = "error: repl.swift:131:8: value of type 'Point<Float>'\n 130>"
    assert repl_output.search_for_error(text) is not None

    text = "var i = 1.0\ni: Double = 1\n 110>"
    assert repl_output.search_for_error(text) is None

    text = "$E11: repl.swift:131:8: value of type 'Point<Float>'\n 130>"
    assert repl_output.search_for_error(text) is not None


def test__strip_prompt_input_lines() -> None:
    output = """
    169>
     170. var point1 = Point<Float>(x: 5, y: 2)
     174. result.translate(dx: 1, dy: 100)
     175.
    $R9: Point<Float> = {
      x = 12
      y = 106
    }
    """
    output = repl_output.remove_prompt_input_lines(output)
    expected_output = """
    $R9: Point<Float> = {
      x = 12
      y = 106
    }
    """
    assert_lines_equal(output, expected_output)


def test__strip_prompt_input_lines__no_inputs() -> None:
    output = """
    $R9: Point<Float> = {
      x = 12
      y = 106
    }
    """
    output = repl_output.remove_prompt_input_lines(output)
    expected_output = """
    $R9: Point<Float> = {
      x = 12
      y = 106
    }
    """
    assert_lines_equal(output, expected_output)


def test__find_variables() -> None:
    output = """
    169>
     170. var point1 = Point<Float>(x: 5, y: 2)
     174. result.translate(dx: 1, dy: 100)
     175. fake: String = "hello"
    pairs: [Tuple<Point, Point>] = 5 values {
    }
    a: String = "hello"
    $R9: Point<Float> = {
      x = 12
      y = 106
    }
    result: Point<Float> = {
      x = 11
      y = 6
    }
    results: [Point<Float>] = 3 values {
      [0] = {
        x = 11
        y = 6
      }
      [1] = {
        x = 5
        y = 2
      }
      [2] = {
        x = 5
        y = 2
      }
    }
    """
    variables = repl_output.find_variables(output)
    exp_variables = {
        "pairs": "[Tuple<Point, Point>]",
        "$R9": "Point<Float>",
        "a": "String",
        "result": "Point<Float>",
        "results": "[Point<Float>]",
    }
    assert len(variables) == len(exp_variables)
    for key, dtype in exp_variables.items():
        assert variables[key][0] == dtype


def test__find_variables__should_ignore_text() -> None:
    output = """
    test: will ignore this line - ok?
    warning: will ignore this line = ok?
    warning: will ignore this line=ok?
    """
    variables = repl_output.find_variables(output)
    assert variables == {}


def test__find_variables__should_parse_correctly() -> None:
    output = """
    a: String = "hello = hello"
    b: [Any] = [1, 2, 3, "hello == hello"]
    """
    variables = repl_output.find_variables(output)
    assert len(variables) == 2
    assert variables["a"] == ("String", '"hello = hello"')
    assert variables["b"] == ("[Any]", '[1, 2, 3, "hello == hello"]')


def test__batch_prompt() -> None:
    batches = repl_output.batch_prompt("let x = 1")
    assert batches == ["let x = 1"]
    prompt = """
    let x = 1
    // comment

    let y = 2
    """
    prompt = "\n".join([line.strip() for line in prompt.split("\n")])
    batches = repl_output.batch_prompt(prompt)
    assert batches == ['let x = 1\nlet y = 2']
    batches = repl_output.batch_prompt(prompt, maxsize=10)
    assert batches == ['let x = 1', 'let y = 2']
    batches = repl_output.batch_prompt(prompt, maxsize=25)
    assert batches == ['let x = 1\nlet y = 2']

def test__split_prompt() -> None:
    blocks = repl_output.split_prompt("let x = 1")
    assert blocks == ["let x = 1"]
    prompt = """
    let x = 1
    // comment

    let y = 2
    """
    prompt = "\n".join([line.strip() for line in prompt.split("\n")])
    blocks = repl_output.split_prompt(prompt)
    assert blocks == [prompt]
    blocks = repl_output.split_prompt(prompt, maxsize=10)
    assert blocks == ['\nlet x = 1', '\n// commen', 't\n\nlet y =', ' 2\n']
    assert [len(b) for b in blocks] == [10, 10, 10, 3]


"""Functions related to parsing Swift code."""
import re
from dataclasses import dataclass
from typing import Self, final


def get_files_content(paths: list[str]) -> str:
    """Get the content of files at the given paths."""
    include_content = []
    for file_path in paths:
        include_code = get_file_content(file_path)
        include_content.append(include_code)
    return "\n".join(include_content)


def get_file_content(path: str) -> str:
    """Get the content of the file at the given path."""
    with open(path) as file:
        file_content = file.read()
    return file_content


@dataclass
class FunctionCode:
    name: str
    code: str
    code_start_line: int
    code_end_line: int
    header: str
    header_start_line: int
    header_end_line: int
    body: str
    body_start_line: int
    body_end_line: int

    @property
    def num_lines(self) -> int:
        return self.code_end_line - self.code_start_line + 1


def make_body_return_var(function_body: str, name: str = "__return_value") -> str:
    lines = function_body.split("\n")
    last_line = lines[-1]
    if last_line.strip().startswith("return"):
        return function_body

    first_line = lines[0]
    match = re.match(r"^\s*", first_line)
    indent = match.group() if match else ""

    # allocate new variable
    new_line = f"{indent}let {name} = {first_line.strip()}"
    lines[0] = new_line
    lines.append(f"{indent}return {name}")
    return "\n".join(lines)


def _count_delta(text: str, start: str, end: str) -> int:
    """Count the difference between the number of occurrences of start and end."""
    return text.count(start) - text.count(end)


def find_function(function_name: str, source_code: str) -> FunctionCode:
    """Extract the code for a function from the given source code text.

    Returns:
        - full function content as a string
        - function body as a string
        - start index of the first body line (0-based)
        - end index of the last body line (0-based)
    """
    lines = source_code.split("\n")

    func_lines = []  # To collect the full function content
    brace_count = 0  # To track opening and closing braces
    inside_function = False  # Flag to indicate if we're inside the target function

    func_start_index = None  # Line where the function definition starts
    body_start_index = None  # Line where the function body starts
    body_end_index = None  # Line where the function body ends

    # Regular expression to match the function definition (including multiline)
    func_def_pattern = re.compile(r"\bfunc\b[\s\S]*?\b" + re.escape(function_name) + r"\b")
    lines_wo_comments = [line.split("//")[0] for line in lines]

    i = 0
    while i < len(lines):
        line = lines[i]
        line_wo_comments = lines_wo_comments[i]

        # If we're not yet inside the function
        if not inside_function:
            # Check if the function definition starts here
            # Collect lines until we find the opening brace '{'
            potential_def_lines = []
            def_start_index = i
            while i < len(lines):
                potential_def_lines.append(lines[i])
                joined_lines = "\n".join(potential_def_lines)
                if re.search(func_def_pattern, joined_lines):
                    if "{" in lines_wo_comments[i]:
                        # Found the opening brace
                        func_start_index = def_start_index
                        func_lines.extend(potential_def_lines)
                        brace_count += _count_delta(lines_wo_comments[i], "{", "}")
                        # Determine where the body starts
                        if lines_wo_comments[i].strip().endswith("{"):
                            body_start_index = i + 1
                        else:
                            body_start_index = i
                        inside_function = True
                        i += 1
                        break
                    else:
                        i += 1
                else:
                    # Not the function we're looking for
                    i += 1
                    break
            else:
                # End of file reached without finding function
                break
        else:
            # We're inside the function body
            func_lines.append(line)
            brace_count += _count_delta(line_wo_comments, "{", "}")
            if brace_count == 0:
                # Function body ends here
                body_end_index = i
                break
            i += 1

    if func_start_index is not None and body_start_index is not None and body_end_index is not None:
        full_function_content = "\n".join(func_lines)
        # The body is from body_start_index to body_end_index - 1
        function_body = "\n".join(lines[body_start_index:body_end_index])
        function_header = "\n".join(lines[func_start_index:body_start_index])
        return FunctionCode(
            name=function_name,
            # full function content
            code=full_function_content,
            code_start_line=func_start_index,
            code_end_line=i,
            # header
            header=function_header,
            header_start_line=func_start_index,
            header_end_line=body_start_index - 1,
            # body
            body=function_body,
            body_start_line=body_start_index,
            body_end_line=body_end_index - 1,
        )
    else:
        raise ValueError(
            f"Function '{function_name}' not found or improperly formatted, check for unmatched "
            f"{{}} or () or [] in the text or comments inside the function."
        )


@final
class CodeBlock:
    def __init__(self, code_lines: list[str], start_line: int, end_line: int):
        self.code_lines = code_lines
        self.start_line = start_line
        self.end_line = end_line

    @property
    def num_lines(self) -> int:
        return len(self.code_lines)

    @property
    def text(self) -> str:
        return "\n".join(self.code_lines)

    def is_comment_block(self) -> bool:
        """Check if each line in the block is a single line comment."""
        for line in self.code_lines:
            if not line.strip().startswith("//"):
                return False
        return True

    def is_function_call_block(self) -> bool:
        line = self.code_lines[0].strip()
        # check for "for (condition text) {" or "while (condition text) {"
        if re.search(r"\s*(for|while)\s*.*{\s*", line):
            return False
        if re.search(r"\s*\w+\s*\(", line):
            return True
        return False

    def is_if_else_block(self) -> bool:
        if not re.search(r"\s*if\s*[(]?", self.code_lines[0]):
            return False

        for line in self.code_lines:
            if re.search(r"\s*else\s*", line):
                print(f"WARNING! If else is not supported: {self.text}")
                return True
        return False

    def can_split(self) -> bool:
        conditions = [
            self.num_lines > 1,
            not self.is_comment_block(),
            not self.is_function_call_block(),
            not self.is_if_else_block(),
        ]
        return all(conditions)

    def split(self) -> list[Self]:
        if not self.can_split():
            raise ValueError("Cannot split this block.")

        inner_blocks = extract_code_blocks(self.code_lines[1:-1])
        # map local lines to global lines
        for block in inner_blocks:
            block.start_line += self.start_line + 1
            block.end_line += self.start_line + 1
        return inner_blocks


def extract_code_blocks(source_lines: list[str]) -> list[CodeBlock]:
    blocks = []
    current_block: list[str] = []
    start_line_num = 0
    grouping_levels = {"paren": 0, "bracket": 0, "brace": 0}

    for i in range(0, len(source_lines)):
        line = source_lines[i]
        if not current_block:
            # Start of a new block
            start_line_num = i
        current_block.append(line)

        # Remove comments
        line_no_comment = line.split("//")[0]
        line_no_strings = _remove_string_literals(line_no_comment)

        # Update grouping levels
        grouping_levels["paren"] += _count_delta(line_no_strings, "(", ")")
        grouping_levels["brace"] += _count_delta(line_no_strings, "{", "}")
        grouping_levels["bracket"] += _count_delta(line_no_strings, "[", "]")

        # Check if grouping levels are zero
        if (
            grouping_levels["paren"] == 0
            and grouping_levels["brace"] == 0
            and grouping_levels["bracket"] == 0
        ):
            # Statement is complete
            end_line_num = i
            blocks.append(
                CodeBlock(
                    code_lines=current_block,
                    start_line=start_line_num,
                    end_line=end_line_num,
                )
            )
            current_block = []
            grouping_levels = {"paren": 0, "bracket": 0, "brace": 0}

    if current_block:
        # Append last block
        end_line_num = len(source_lines) - 1
        blocks.append(
            CodeBlock(
                code_lines=current_block,
                start_line=start_line_num,
                end_line=end_line_num,
            )
        )
    # remove empty blocks
    blocks = [block for block in blocks if block.text.strip() != ""]
    return blocks


def _remove_string_literals(text: str) -> str:
    """Remove string literals to avoid counting braces inside strings"""
    result = ""
    in_string = False
    escape_next = False
    for c in text:
        if escape_next:
            escape_next = False
            continue
        if c == "\\":
            escape_next = True
            continue
        if c == '"' and not in_string:
            in_string = True
            continue
        elif c == '"' and in_string:
            in_string = False
            continue
        if not in_string:
            result += c
    return result



INIT_COMMANDS = """
import Foundation
import Dispatch

/// Function to deserialize an object from a JSON file at the given path
func _deserializeObject<T: Decodable>(_ path: String) throws -> T {
    let url = URL(fileURLWithPath: path)
    let data = try Data(contentsOf: url)
    let decoder = JSONDecoder()
    let object = try decoder.decode(T.self, from: data)
    return object
}

/// Function to serialize an object and save it as a JSON file at the given path
func _serializeObject<T: Encodable>(_ object: T, to path: String) throws {
    let encoder = JSONEncoder()
    encoder.outputFormatting = [.prettyPrinted, .sortedKeys] // Optional formatting
    let data = try encoder.encode(object)
    let url = URL(fileURLWithPath: path)
    try data.write(to: url)
}

/// Runs async function in a synchronous manner. REPL crashes when await is called in the
/// main thread.
func runSync<T>(_ asyncClosure: @escaping () async throws -> T) throws -> T {
    let semaphore = DispatchSemaphore(value: 0)
    var result: Result<T, Error>!

    Task {
        do {
            let value = try await asyncClosure()
            result = .success(value)
        } catch {
            result = .failure(error)
        }
        semaphore.signal()
    }

    semaphore.wait()
    return try result.get()
}

"""

END_OF_INCLUDE = "// -- END OF AUTO REPL INCLUDE --"



import json
import os
import re
import tempfile
from dataclasses import dataclass
from pathlib import Path
from typing import Any
import sys
import pexpect

from repltilian import code, constants, profiler, repl_output


class SwiftREPLException(Exception):
    pass


@dataclass
class Options:
    output_hide_inputs: bool = True
    output_hide_variables: bool = False
    output_stop_pattern: str | None = None
    timeout: float = 0.01
    maxread: int = 4096
    maxsend: int = 1000 if sys.platform == "darwin" else 2000


class SwiftREPL:
    def __init__(self, cwd: str | None = None, options: Options = Options()) -> None:
        """Initialize the Swift REPL.

        Args:
            cwd: optional path to the working directory, a folder with Package.swift file. If
                provided, the REPL will be started with `swift run --repl` command, otherwise
                with `swift repl`.
            options: an instance of REPLOptions class with optional parameters for REPL output.
        """
        self.cwd = cwd
        self.options = options
        self.vars = VariablesRegister(self)
        self._initialized = False
        self._reload_paths: set[str] = set()
        self._output: str | None = None

        self._process = self._initiate_repl()
        self.run(constants.INIT_COMMANDS, verbose=False)
        self.run("""print("REPL is running !")""")

    def _initiate_repl(self) -> pexpect.spawn:
        env = os.environ.copy()
        env = {"PATH": env["PATH"], "SHELL": env["SHELL"], "TERM": "dumb"}
        command = "swift run --repl"
        if self.cwd is None:
            command = "swift repl"

        self._process = pexpect.spawn(
            command=command,
            encoding="utf-8",
            timeout=1,
            env=env,
            cwd=self.cwd,
        )
        self._initialized = True
        return self._process

    def add_reload_file(self, path: str | Path) -> None:
        """Path to file which will be added to the REPL input before running the code."""
        if path not in self._reload_paths:
            if not Path(path).is_file():
                raise FileNotFoundError(f"File '{path}' does not exist.")
            self._reload_paths.add(str(path))

    def clear_reload_files(self) -> None:
        """Clear the list of files which are reloaded before running the code."""
        self._reload_paths.clear()

    def run(
        self,
        prompt: str,
        autoreload: bool = False,
        verbose: bool = True,
    ) -> None:
        if not self._initialized:
            raise SwiftREPLException("REPL is not initialized.")

        include_paths: list[str] | None = None
        if self._reload_paths and autoreload:
            include_paths = list(self._reload_paths)
        if include_paths:
            include_text = code.get_files_content(include_paths)
            prompt = include_text + "\n" + constants.END_OF_INCLUDE + "\n" + prompt

        if not prompt.startswith("\n"):
            prompt = "\n" + prompt

        blocks = repl_output.batch_prompt(prompt, self.options.maxsend)
        repl_raw_outputs = []
        while blocks:
            block = blocks.pop(0)
            self._process.sendline(block)
            while True:
                try:
                    buffer = self._process.read_nonblocking(
                        size=self.options.maxread,
                        timeout=self.options.timeout,
                    )
                    repl_raw_outputs.append(buffer)
                except pexpect.exceptions.EOF as e:
                    raise SwiftREPLException(
                        f"REPL crashed with error: '{e}'. Did you try to run "
                        f"async function ? If yes consider to use: 'try runSync "
                        f"{{ try await yourAsyncFunction }}'"
                    )
                except pexpect.exceptions.TIMEOUT:
                    if blocks:
                        break
                    # a regex which matches the waiting prompt e.g. "1>" or "102>" but
                    # there must not be any text after the prompt
                    prompt_pattern = re.compile(r"(\d+>$)")
                    buffer_end = "".join(repl_raw_outputs[-10:])
                    has_prompt = prompt_pattern.search(repl_output.clean(buffer_end))
                    if has_prompt is None:
                        continue
                    break
                except Exception as e:
                    raise SwiftREPLException(f"REPL error: {e}")

        output = repl_output.clean("".join(repl_raw_outputs))
        self._output = output
        if error_line := repl_output.search_for_error(output):
            repl_output.print_output(output)
            raise SwiftREPLException(f"Error in Swift code: '{error_line}'")

        if verbose:
            repl_output.print_output(
                output,
                stop_output_at_pattern=self.options.output_stop_pattern,
                hide_inputs=self.options.output_hide_inputs,
                hide_variables=self.options.output_hide_variables,
            )

        variable_updates = repl_output.find_variables(output)
        for key, (dtype, value) in variable_updates.items():
            self.vars[key] = Variable(self, key, dtype, value)

    def line_profile(
        self,
        prompt: str,
        function_name: str,
        source_path: str,
        autoreload: bool = False,
    ) -> None:
        """Return the code with line profiling instrumentation."""
        source_code = code.get_file_content(source_path)
        profiled_function = profiler.get_function_for_line_profiler(function_name, source_code)
        prompt = profiled_function + "\n" + prompt
        self.run(prompt, autoreload=autoreload)

    def close(self) -> None:
        self._process.sendline(":quit")
        self._process.terminate()
        self._process.close()
        self._initialized = False


class Variable:
    def __init__(
        self,
        repl_ref: SwiftREPL,
        name: str,
        dtype: str | None = None,
        value: str | None = None,
    ):
        self.name = name
        self.dtype = dtype
        self.value = value
        self._repl = repl_ref

    def get(self, verbose: bool = False) -> Any:
        """Return the JSON representation of the variable obtained through
        the JSON deserialization from REPL process.
        """
        if self._repl is None:
            raise SwiftREPLException("Variable is not associated with a REPL instance.")

        with tempfile.NamedTemporaryFile() as tmpfile:
            path = f"{tmpfile.name}.json"
            self._repl.run(
                f'_serializeObject({self.name}, to: "{path}")',
                verbose=verbose,
                autoreload=False,
            )
            with open(path) as file:
                data = json.load(file)
            return data

    def __repr__(self) -> str:
        return f"{self.name}[{self.dtype}] at {id(self)}"


class VariablesRegister(dict[str, Variable]):
    """A class which is responsible for managing variables in the REPL."""

    def __init__(self, repl_ref: SwiftREPL) -> None:
        super().__init__()
        self._repl_ref = repl_ref

    def __setitem__(self, key: str, value: Any | Variable) -> None:
        if not isinstance(value, Variable):
            raise SwiftREPLException(
                "Only Variable instances can be added to the register, use set " "method instead."
            )
        super().__setitem__(key, value)

    def __getitem__(self, key: str) -> Variable:
        if key not in self:
            return Variable(self._repl_ref, key)
        return super().__getitem__(key)

    def set(self, name: str, dtype: str, value: Any, verbose: bool = False) -> None:
        """Set a variable in the REPL with the given name, type and value. This function will
        create or update existing variable.
        """
        with tempfile.NamedTemporaryFile() as tmpfile:
            path = f"{tmpfile.name}.json"
            with open(path, "w") as fp:
                json.dump(value, fp)

            self._repl_ref.run(
                f'\nvar {name}: {dtype} = try _deserializeObject("{path}")\n',
                verbose=verbose,
                autoreload=False,
            )
            self[name] = Variable(self._repl_ref, name, dtype, value)



"""Functions related to line profiler functionality."""
import re

from repltilian import code


def get_function_for_line_profiler(function_name: str, source_code: str) -> str:
    function = code.find_function(function_name, source_code)
    body_lines = code.make_body_return_var(function.body).split("\n")
    body_lines = [line for line in body_lines if line.strip()]
    return_line, body_lines = body_lines[-1], body_lines[:-1]

    # Initialize the profiling variables
    indent_match = re.match(r"\s*", body_lines[0])
    indent = indent_match.group() if indent_match else ""
    num_lines = len(body_lines)

    line_contents: dict[int, str] = {}
    instrumented_lines: list[str] = [
        indent + "var __line_times = [Int: UInt64]()",
        indent + "var __line_hits = [Int: Int]()",
        indent + f"(0..<{num_lines}).map {{ i in __line_times[i] = 0}}",
        indent + f"(0..<{num_lines}).map {{ i in __line_hits[i] = 0}}",
        indent + "let __start_time_func = DispatchTime.now().uptimeNanoseconds",
    ]

    blocks = code.extract_code_blocks(body_lines)
    for block in blocks:
        render_for_profile(block, instrumented_lines, line_contents)

    instrumented_lines.append(indent + "let __end_time_func = DispatchTime.now().uptimeNanoseconds")

    # Insert code to print profiling statistics before the closing brace
    profiling_output = [
        indent + 'print("Timer unit: 1 ns")',
        indent + "let __total_time = __end_time_func - __start_time_func",
        indent + 'print(String(format: "\\nTotal time: %.3f s", Double('
        "__total_time)/1_000_000_000))",
        indent + f'print("Function: {function_name} at line {function.code_start_line + 1}")',
        indent + 'print("")',
        indent + 'print("Line #      Hits         Time   Per Hit   % Time  Line Contents")',
        indent + 'print("===============================================================")',
        indent + "for line in (__line_times.keys.sorted()) {",
        indent + "    let hits = __line_hits[line] ?? 0",
        indent + "    let time = __line_times[line] ?? 0",
        indent + "    let per_hit = hits > 0 ? Double(time) / Double(hits) : 0",
        indent + "    let percent_time = __total_time > 0 "
        "? (Double(time) / Double(__total_time)) * 100 : 0",
        indent + "    let lineContentDict: [Int: String] = [",
    ]

    # Add line contents to the dictionary
    for ln in sorted(line_contents.keys()):
        content = line_contents[ln]
        profiling_output.append(indent + f'        {ln}: "{content}",')
    profiling_output.append(indent + "    ]")
    profiling_output.append(indent + '    let contents = lineContentDict[line] ?? ""')
    profiling_output.append(
        indent + '    print(String(format: "%6d %10d %12.6f %9.6f %8.1f%%  %@", line, hits, '
        "Double(time)/1_000_000_000, per_hit/1_000_000_000, percent_time, contents))"
    )
    profiling_output.append(indent + "}")
    # Find the closing brace line and insert profiling output before it
    instrumented_lines = instrumented_lines + profiling_output + [return_line]
    match = re.match(r"\s*", function.header)
    header_intent = match.group() if match else ""
    instrumented_lines = [function.header] + instrumented_lines + [header_intent + "}"]

    return "\n".join(instrumented_lines)


def render_for_profile(
    block: code.CodeBlock,
    instrumented_lines: list[str],
    line_contents: dict[int, str],
) -> None:
    if block.num_lines == 0:
        return

    lines = block.code_lines
    match = re.match(r"\s*", lines[0])
    indent = match.group() if match else ""
    line_number = block.start_line

    start_time_var = f"__start_time_{line_number}"
    end_time_var = f"__end_time_{line_number}"
    # Insert timing code
    instrumented_lines.append(
        indent + f"let {start_time_var} = DispatchTime.now().uptimeNanoseconds"
    )
    if len(lines) == 1:
        line_contents[line_number] = lines[0].replace('"', '\\"')
        instrumented_lines.append(lines[0])
    elif not block.can_split():
        for i, line in enumerate(lines):
            line_contents[line_number + i] = line.replace('"', '\\"')
            instrumented_lines.append(line)
    else:
        instrumented_lines.append(lines[0])
        line_contents[line_number] = lines[0].replace('"', '\\"')
        for inner in block.split():
            render_for_profile(inner, instrumented_lines, line_contents)
        instrumented_lines.append(lines[-1])
        line_contents[block.end_line] = lines[-1].replace('"', '\\"')

    instrumented_lines.append(indent + f"let {end_time_var} = DispatchTime.now().uptimeNanoseconds")
    instrumented_lines.append(
        indent + f"__line_times[{line_number}] = (__line_times[{line_number}] ?? 0) "
        f"+ ({end_time_var} - {start_time_var})"
    )
    instrumented_lines.append(
        indent + f"__line_hits[{line_number}] = (__line_hits[{line_number}] ?? 0) + 1"
    )



import pytest

from repltilian import SwiftREPL, SwiftREPLException


def test_add_reload_file(repl: SwiftREPL, sample_filepath: str) -> None:
    repl.add_reload_file(sample_filepath)
    assert repl._reload_paths == {sample_filepath}
    repl.add_reload_file(sample_filepath)
    assert repl._reload_paths == {sample_filepath}
    repl.clear_reload_files()
    assert repl._reload_paths == set()


def test__allocate_variable(repl: SwiftREPL) -> None:
    repl.run("let x = 5")
    assert repl.vars["x"].get() == 5

    repl.vars.set("x", "Int", 10)
    assert repl.vars["x"].get() == 10

    repl.vars.set("array", "Array<Int>", [1, 2, 3])
    assert repl.vars["array"].get() == [1, 2, 3]


def test__create_variable_from_included_file(repl: SwiftREPL, sample_filepath: str) -> None:
    with pytest.raises(SwiftREPLException):
        repl.run("var point = Point<Float>(x: 1, y: 2)")

    repl.add_reload_file(sample_filepath)
    with pytest.raises(SwiftREPLException):
        repl.run("var point = Point<Float>(x: 1, y: 2)")

    repl.run("var point = Point<Float>(x: 1, y: 2)", autoreload=True)

    point_data = repl.vars["point"].get()
    assert point_data == {"x": 1, "y": 2}


def test__get_variable__not_in_register(repl: SwiftREPL) -> None:
    repl.run("let x = 5")
    assert repl.vars["x"].get() == 5

    del repl.vars["x"]
    assert repl.vars["x"].get() == 5


def test__run_async_function__should_fail_on_await(repl: SwiftREPL) -> None:
    repl.run(
        """
    func sum(_ a: Int, _ b: Int) async -> Int {
        return a + b
    }
    """
    )
    with pytest.raises(SwiftREPLException):
        repl.run("let result = await sum(5, 7)")


def test__run_async_function(repl: SwiftREPL) -> None:
    repl.run(
        """
    func sum(_ a: Int, _ b: Int) async -> Int {
        return a + b
    }
    """
    )
    repl.run("let result = try runSync {await sum(5, 7)}")
    assert repl.vars["result"].get() == 12



from repltilian.repl import SwiftREPL, SwiftREPLException  # noqa: F401

__all__ = ["SwiftREPL", "SwiftREPLException", "load_ipython_extension"]


def load_ipython_extension(ipython):  # type: ignore
    from repltilian.ipython import REPLMagic

    magic = REPLMagic(ipython)
    ipython.register_magics(magic)



import pytest

from repltilian import code


def test__find_function__simple_function(sample_code: str) -> None:
    function = code.find_function("translate", sample_code)

    assert function.name == "translate"
    assert function.body.strip() == "return Point(x: x + dx, y: y + dy)"
    assert function.header.strip() == "func translate(dx: T, dy: T) -> Point<T> {"
    assert function.body_start_line == function.body_end_line
    assert function.header_start_line == function.header_end_line
    assert function.body_start_line == function.header_end_line + 1
    assert function.num_lines == 3


def test_find_function__should_raise_error(sample_code: str) -> None:
    with pytest.raises(ValueError):
        code.find_function("trans", sample_code)


def test__find_function__multiline_function(sample_code: str) -> None:
    function = code.find_function("findKNearestNeighbors", sample_code)

    assert function.name == "findKNearestNeighbors"
    assert function.header.strip() == (
        "func findKNearestNeighbors<T: FloatingPoint>(query: ["
        "Point<T>], dataset: [Point<T>], k: Int = 5) -> [SearchResult<T>] {"
    )
    assert function.header_start_line == function.code_start_line
    assert function.header_start_line == function.header_end_line
    assert function.body_start_line == function.header_end_line + 1
    assert function.code_end_line == function.body_end_line + 1
    assert function.num_lines == 27


def test__extract_code_blocks__simple_function(sample_code: str) -> None:
    function = code.find_function("translate", sample_code)
    blocks = code.extract_code_blocks(function.body.split("\n"))

    assert len(blocks) == 1
    assert blocks[0].text.strip() == "return Point(x: x + dx, y: y + dy)"


def test__extract_code_blocks__multiline_function(sample_code: str) -> None:
    function = code.find_function("findKNearestNeighbors", sample_code)
    blocks = code.extract_code_blocks(function.body.split("\n"))

    assert len(blocks) == 3
    assert blocks[0].text.strip() == "var results: [SearchResult<T>] = []"
    assert blocks[1].code_lines[0].strip() == "for queryPoint in query {"
    assert blocks[2].code_lines[0].strip() == "return results"

    with pytest.raises(ValueError):
        blocks[0].split()


def test__extract_code_blocks__function_has_string_with_brackets(sample_code: str) -> None:
    function = code.find_function("removeBrackets", sample_code)
    blocks = code.extract_code_blocks(function.body.split("\n"))

    assert len(blocks) == 2
    assert blocks[0].text.strip() == 'let brackets = ["[", "]"]'
    assert blocks[1].text.strip() == "return text.filter { !brackets.contains(String($0)) }"


def test__make_body_return_var(sample_code: str) -> None:
    body = """return Point(x: x + dx, y: y + dy)"""
    new_body = code.make_body_return_var(body)
    assert body == new_body
    body = """Point(x: x + dx, y: y + dy)"""
    new_body = code.make_body_return_var(body, "val")
    assert new_body == "let val = Point(x: x + dx, y: y + dy)\nreturn val"



"""Python setup.py for repltilian package."""

import os

from setuptools import find_packages, setup


def read(filepath: str) -> str:
    with open(
        os.path.join(os.path.dirname(__file__), filepath),
        encoding="utf-8",
    ) as open_file:
        content = open_file.read().strip()
    return content


def read_requirements(path: str) -> list[str]:
    return [
        line.strip()
        for line in read(path).split("\n")
        if not line.startswith(('"', "#", "-", "git+"))
    ]


setup(
    name="repltilian",
    version=read("repltilian/VERSION"),
    description="Awesome repltilian created by kmkolasinski",
    url="https://github.com/kmkolasinski/repltilian/",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    author="kmkolasinski",
    packages=find_packages(exclude=["tests", ".github", "notebooks"]),
    install_requires=read_requirements("requirements.txt"),
    extras_require={"test": read_requirements("requirements-test.txt")},
)
